#!/usr/bin/env python

"""
A processing framework for iMIP content.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from email import message_from_file
from imiptools.content import handle_itip_part
from imiptools.data import get_addresses, get_uri
from imiptools.mail import Messenger
from imiptools.profile import Preferences
import sys

# Postfix exit codes.

EX_TEMPFAIL     = 75

# Permitted iTIP content types.

itip_content_types = [
    "text/calendar",                        # from RFC 6047
    "text/x-vcalendar", "application/ics",  # other possibilities
    ]

# Processing of incoming messages.

def get_all_values(msg, key):
    l = []
    for v in msg.get_all(key) or []:
        l += [s.strip() for s in v.split(",")]
    return l

class Processor:

    "The processing framework."

    def __init__(self, handlers, messenger=None):
        self.handlers = handlers
        self.messenger = messenger or Messenger()
        self.lmtp_socket = None

    def process(self, f, original_recipients, recipients, outgoing_only):

        """
        Process content from the stream 'f' accompanied by the given
        'original_recipients' and 'recipients'.
        """

        msg = message_from_file(f)
        senders = get_addresses(msg.get_all("Reply-To") or msg.get_all("From") or [])
        original_recipients = original_recipients or get_addresses(get_all_values(msg, "To") or [])

        # Handle messages with iTIP parts.

        for recipient in original_recipients:
            self.process_for_recipient(msg, recipient, senders, outgoing_only)

    def process_for_recipient(self, msg, recipient, senders, outgoing_only):

        """
        Process the given 'msg' for a single 'recipient', having the given
        'senders', and with the given 'outgoing_only' status.

        Processing individually means that contributions to resulting messages
        may be constructed according to individual preferences.
        """

        handlers = dict([(name, cls(senders, recipient, self.messenger)) for name, cls in self.handlers])
        handled = False

        for part in msg.walk():
            if part.get_content_type() in itip_content_types and \
               part.get_param("method"):

                handle_itip_part(part, handlers)
                handled = True

        # When processing outgoing messages, no replies or deliveries are
        # performed.

        if outgoing_only:
            return

        # Get responses from the handlers.

        all_responses = []
        for handler in handlers.values():
            all_responses += handler.get_results()

        # Pack any returned parts into messages.

        if all_responses:
            outgoing_parts = {}
            forwarded_parts = []

            for outgoing_recipients, part in all_responses:
                if outgoing_recipients:
                    for outgoing_recipient in outgoing_recipients:
                        if not outgoing_parts.has_key(outgoing_recipient):
                            outgoing_parts[outgoing_recipient] = []
                        outgoing_parts[outgoing_recipient].append(part)
                else:
                    forwarded_parts.append(part)

            # Reply using any outgoing parts in a new message.

            if outgoing_parts:
                for outgoing_recipient, parts in outgoing_parts.items():
                    message = self.messenger.make_outgoing_message(parts, [outgoing_recipient])

                    if "-d" in sys.argv:
                        print >>sys.stderr, "Outgoing parts for %s..." % outgoing_recipient
                        print message
                    else:
                        self.messenger.sendmail([outgoing_recipient], message.as_string())

            # Forward messages to their recipients either wrapping the existing
            # message, accompanying it or replacing it.

            if forwarded_parts:

                # Determine whether to wrap, accompany or replace the message.

                preferences = Preferences(get_uri(recipient))

                incoming = preferences.get("incoming")

                if incoming == "message-only":
                    messages = [msg]
                else:
                    summary = self.messenger.make_summary_message(msg, forwarded_parts)
                    if incoming == "summary-then-message":
                        messages = [summary, msg]
                    elif incoming == "message-then-summary":
                        messages = [msg, summary]
                    elif incoming == "summary-only":
                        messages = [summary]
                    else: # incoming == "summary-wraps-message":
                        messages = [self.messenger.wrap_message(msg, forwarded_parts)]

                for message in messages:
                    if "-d" in sys.argv:
                        print >>sys.stderr, "Forwarded parts..."
                        print message
                    elif self.lmtp_socket:
                        self.messenger.sendmail(recipient, message.as_string(), lmtp_socket=self.lmtp_socket)

        # Unhandled messages are delivered as they are.

        if not handled:
            if "-d" in sys.argv:
                print >>sys.stderr, "Unhandled parts..."
                print msg
            elif self.lmtp_socket:
                self.messenger.sendmail(recipient, msg.as_string(), lmtp_socket=self.lmtp_socket)

    def process_args(self, args, stream):

        """
        Interpret the given program arguments 'args' and process input from the
        given 'stream'.
        """

        # Obtain the different kinds of recipients plus sender address.

        original_recipients = []
        recipients = []
        senders = []
        lmtp = []
        outgoing_only = False

        l = []

        for arg in args:

            # Detect outgoing processing mode.

            if arg == "-O":
                outgoing_only = True

            # Switch to collecting recipients.

            if arg == "-o":
                l = original_recipients
            elif arg == "-r":
                l = recipients

            # Switch to collecting senders.

            elif arg == "-s":
                l = senders

            # Switch to getting the LMTP socket.

            elif arg == "-l":
                l = lmtp

            # Ignore debugging options.

            elif arg == "-d":
                pass
            else:
                l.append(arg)

        self.messenger.sender = senders and senders[0] or self.messenger.sender
        self.lmtp_socket = lmtp and lmtp[0] or None
        self.process(stream, original_recipients, recipients, outgoing_only)

    def __call__(self):

        """
        Obtain arguments from the command line to initialise the processor
        before invoking it.
        """

        args = sys.argv[1:]

        if "-d" in args:
            self.process_args(args, sys.stdin)
        else:
            try:
                self.process_args(args, sys.stdin)
            except SystemExit, value:
                sys.exit(value)
            except Exception, exc:
                if "-v" in args:
                    raise
                type, value, tb = sys.exc_info()
                print >>sys.stderr, "Exception %s at %d" % (exc, tb.tb_lineno)
                #import traceback
                #traceback.print_exc(file=open("/tmp/mail.log", "a"))
                sys.exit(EX_TEMPFAIL)
        sys.exit(0)

# vim: tabstop=4 expandtab shiftwidth=4
